/*
  Lab Solutions for Lab 7 Tasks 1-2
  
  Created by Michael Bykov (michael.bykov@wsu.edu)
  
  Discussion of the assigned questions.
*/


#include "List.h"

/*

	Shallow vs Deep Copy:
	
	When we are copying an object that manages dynamically allocated memory, there are two distinct ways we can copy said object:
	- A shallow copy of an object copies the contents of all of the object's fields directly.
		When looking at pointers specifically, this means that the address stored in the pointer was copied but no new data was allocated, so after a shallow copy both
		the original and the copy of the object will point at the same resource.
		This can be extremely problematic when, for example, one of the objects deletes the memory that it points to (for instance, when its destructor is called) but the other object is still pointing at the same memory, resulting in a dangling pointer.
		Outside of memory errors, this can also cause unexpected behavior when you updated one of the objects (for instance, adding an item to the end of a list) and both objects end up having the new data when you only expected the object you changed to have the new data.
	- To counteract these problems, we can perform a deep copy which doesn't just copy the contents of all of the fields blindly, but also allocates new memory for the resources managed by the original object,
		so that the copy of the object doesn't share the same memory as the original object.
		For instance, with a linked list, this means allocating new space for all of the nodes and linking them to the new list.
	
	Here is an example of the difference between a shallow copy and a deep copy with a linked list:
		$200     $300     $400
		[ 5 ] -> [ 6 ] -> [ 7 ] -> NULL
		
		original list: { mpHead = $200 }
		
		shallow copy: { mpHead = $200 } // Both of the lists point to the same linked list
		
		$500     $600     $700
		[ 5 ] -> [ 6 ] -> [ 7 ] -> NULL    // A node-by-node copy is made of the list with the same values, but not the same pointers. None of the pointers in the new list point back into the old list.
		
		deep copy: { mpHead = $500 } // The copy of the list points to a new linked list that is a complete copy (duplication) of the original linked list

*/

int main(void)
{
	// Start with debugging this project! "Step Into" each statement to answer the questions!
	// Question 1: what function is called in the statement below?
	/*
		The ListNode(int data) constructor is called with an argument of 0,
		treating the the constructor as our de facto default constructor.
	*/
	ListNode n1;
	// Question 2: what function is called in the statement below?
	/*
		The same ListNode(int data) constructor is called but now with an argument of 42 rather than 0.
		Since their is no dynamic memory allocated by the int 42, the only copy operation we can perform is a shallow copy.
	*/
	ListNode n2(42); // shallow or deep copy?
	// Question 3: what function is called in the statement below?
	/*
		The copy constructor is invoked which performs a shallow copy since it does not allocate new memory for any of the subsequent nodes it points to, it just copies the pointer to them directly.
	*/
	ListNode n3(n2); // shallow or deep copy?
	// Question 4: what function is called in the statement below?
	/*
		This is the same statement as above just written in a different notation.
		Since we are declaring a new variable and initializing it here, the equals sign doesn't invoke the copy assignment operator but rather serves as a way to write the copy constructor in a different (more natural) way.
	*/
	ListNode n4 = n3; // shallow or deep copy?

	ListNode n5;
	// Question 5: what function/operator is called in the statement below?
	/*
		The copy assignment operator is invoked but since we didn't explicitly define one, the built-in copy assignment operator is used which performs a shallow copy.
	*/
	n5 = n4; // did we explicitly implement this function/operator? 

	// Question 6: what function is called in the statement below?
	/*
		The default constructor is called whenever we don't explicitly call any other constructor, like in this statement.
	*/
	List l1;
	// make sure that you finish implementing the copy constructor for the List object before you try to 
	// execute the next statement!
	// Question 7: what function is called in the statement below? What would happen if we did not 
	//             explicitly implement the function? Would the compiler generated one be good enough?
	/*
		The copy constructor which we implemented as a deep copy.
		If we didn't explicitly define the copy constructor, the compiler would have generated one for us which would have performed a shallow copy.
		A shallow copy would be problematic in this situation because then both l2 and l1 would share ownership of the linked list so when one of the variables changes the list in a way that the head pointer is updated,
		such as when something is inserted at front or when the first node is deleted, the other list is not updated and still points to the old head pointer which is now invalid.
	*/
	List l2 = l1; // shallow or deep copy?
	// Question 8: what would happen if a shallow copy constructor is used, instead of a deep copy constructor, 
	//             when the destructor for each list object is implicily invoked? l2 and l1 would have the same
	//             address stored in the head pointer, and thus, one of the lists would be freed and when the other went
	//             out of scope, the destructor would try to access already freed memory (a dangling pointer situation)!

	List l3;
	// make sure that you finish implementing the overloaded assignment for the List object before you try to 
	// execute the next statement!
	// Question 9: what function is called in the statement below? What would happen if we did not 
	//             explicitly implement the function? Would the compiler generated one be good enough?
	/*
		The copy assignment operator that we overloaded.
		If we didn't explicitly define the copy assignment operator, the compiler would have generated one for us which would have performed a shallow copy.
		A shallow copy would be problematic in this situation in the same way it would be problematic in the copy constructor.
	*/
	l3 = l2; // shallow or deep copy? 
	// Question 10: what would happen if a shallow copy assignment is used, instead of a deep copy assignment, 
	//              when the destructor for each list object is implicily invoked? l3 and l2 would have the same
	//              address stored in the head pointer, and thus, one of the lists would be freed and when the other went
	//              out of scope, the destructor would try to access already freed memory (a dangling pointer situation)!

	// Question 11: how does the List destructor free up the nodes? You will need to visit "List.cpp" to answer this question!
	/*
		Recursively!
		It visits each node one by one and deletes them starting from the last node and working its way back to the first node.
	*/

	return 0;
}
